@@ -30,11 +30,12 @@ use jni::sys::jsize;
 use jni::JNIEnv;
 use opendal::layers::BlockingLayer;
 use opendal::raw::PresignedRequest;
+use opendal::Entry;
 use opendal::Operator;
 use opendal::Scheme;
 
-use crate::convert::jmap_to_hashmap;
-use crate::convert::jstring_to_string;
+use crate::convert::{jmap_to_hashmap, read_map_field, read_string_field};
+use crate::convert::{jstring_to_string, read_bool_field};
 use crate::executor::executor_or_default;
 use crate::executor::get_current_env;
 use crate::executor::Executor;
@@ -110,8 +111,9 @@ pub unsafe extern "system" fn Java_org_apache_opendal_AsyncOperator_write(
     executor: *const Executor,
     path: JString,
     content: JByteArray,
+    write_options: JObject,
 ) -> jlong {
-    intern_write(&mut env, op, executor, path, content).unwrap_or_else(|e| {
+    intern_write(&mut env, op, executor, path, content, write_options).unwrap_or_else(|e| {
         e.throw(&mut env);
         0
     })
@@ -123,25 +125,59 @@ fn intern_write(
     executor: *const Executor,
     path: JString,
     content: JByteArray,
+    options: JObject,
 ) -> Result<jlong> {
     let op = unsafe { &mut *op };
     let id = request_id(env)?;
 
     let path = jstring_to_string(env, &path)?;
     let content = env.convert_byte_array(content)?;
+    let content_type = read_string_field(env, &options, "contentType")?;
+    let content_disposition = read_string_field(env, &options, "contentDisposition")?;
+    let content_encoding = read_string_field(env, &options, "contentEncoding")?;
+    let cache_control = read_string_field(env, &options, "cacheControl")?;
+    let if_match = read_string_field(env, &options, "ifMatch")?;
+    let if_none_match = read_string_field(env, &options, "ifNoneMatch")?;
+    let append = read_bool_field(env, &options, "append")?;
+    let if_not_exists = read_bool_field(env, &options, "ifNotExists")?;
+    let user_metadata = read_map_field(env, &options, "userMetadata")?;
+
+    let mut write_op = op.write_with(&path, content);
+    if let Some(content_type) = content_type {
+        write_op = write_op.content_type(&content_type);
+    }
+    if let Some(content_disposition) = content_disposition {
+        write_op = write_op.content_disposition(&content_disposition);
+    }
+    if let Some(content_encoding) = content_encoding {
+        write_op = write_op.content_encoding(&content_encoding);
+    }
+    if let Some(cache_control) = cache_control {
+        write_op = write_op.cache_control(&cache_control);
+    }
+    if let Some(if_match) = if_match {
+        write_op = write_op.if_match(&if_match);
+    }
+    if let Some(if_none_match) = if_none_match {
+        write_op = write_op.if_none_match(&if_none_match);
+    }
+    if let Some(user_metadata) = user_metadata {
+        write_op = write_op.user_metadata(user_metadata);
+    }
+    write_op = write_op.if_not_exists(if_not_exists);
+    write_op = write_op.append(append);
 
     executor_or_default(env, executor)?.spawn(async move {
-        let result = do_write(op, path, content).await;
-        complete_future(id, result.map(|_| JValueOwned::Void))
+        let result = write_op
+            .await
+            .map(|_| JValueOwned::Void)
+            .map_err(Into::into);
+        complete_future(id, result)
     });
 
     Ok(id)
 }
 
-async fn do_write(op: &mut Operator, path: String, content: Vec<u8>) -> Result<()> {
-    Ok(op.write(&path, content).await.map(|_| ())?)
-}
-
 /// # Safety
 ///
 /// This function should not be called before the Operator is ready.
@@ -174,21 +210,18 @@ fn intern_append(
     let content = env.convert_byte_array(content)?;
 
     executor_or_default(env, executor)?.spawn(async move {
-        let result = do_append(op, path, content).await;
-        complete_future(id, result.map(|_| JValueOwned::Void))
+        let result = op
+            .write_with(&path, content)
+            .append(true)
+            .await
+            .map(|_| JValueOwned::Void)
+            .map_err(Into::into);
+        complete_future(id, result)
     });
 
     Ok(id)
 }
 
-async fn do_append(op: &mut Operator, path: String, content: Vec<u8>) -> Result<()> {
-    Ok(op
-        .write_with(&path, content)
-        .append(true)
-        .await
-        .map(|_| ())?)
-}
-
 /// # Safety
 ///
 /// This function should not be called before the Operator is ready.
@@ -304,17 +337,17 @@ fn intern_delete(
     let path = jstring_to_string(env, &path)?;
 
     executor_or_default(env, executor)?.spawn(async move {
-        let result = do_delete(op, path).await;
-        complete_future(id, result.map(|_| JValueOwned::Void))
+        let result = op
+            .delete(&path)
+            .await
+            .map(|_| JValueOwned::Void)
+            .map_err(Into::into);
+        complete_future(id, result)
     });
 
     Ok(id)
 }
 
-async fn do_delete(op: &mut Operator, path: String) -> Result<()> {
-    Ok(op.delete(&path).await?)
-}
-
 /// # Safety
 ///
 /// This function should not be called before the Operator is ready.
@@ -377,17 +410,17 @@ fn intern_create_dir(
     let path = jstring_to_string(env, &path)?;
 
     executor_or_default(env, executor)?.spawn(async move {
-        let result = do_create_dir(op, path).await;
-        complete_future(id, result.map(|_| JValueOwned::Void))
+        let result = op
+            .create_dir(&path)
+            .await
+            .map(|_| JValueOwned::Void)
+            .map_err(Into::into);
+        complete_future(id, result)
     });
 
     Ok(id)
 }
 
-async fn do_create_dir(op: &mut Operator, path: String) -> Result<()> {
-    Ok(op.create_dir(&path).await?)
-}
-
 /// # Safety
 ///
 /// This function should not be called before the Operator is ready.
@@ -420,17 +453,17 @@ fn intern_copy(
     let target_path = jstring_to_string(env, &target_path)?;
 
     executor_or_default(env, executor)?.spawn(async move {
-        let result = do_copy(op, source_path, target_path).await;
-        complete_future(id, result.map(|_| JValueOwned::Void))
+        let result = op
+            .copy(&source_path, &target_path)
+            .await
+            .map(|_| JValueOwned::Void)
+            .map_err(Into::into);
+        complete_future(id, result)
     });
 
     Ok(id)
 }
 
-async fn do_copy(op: &mut Operator, source_path: String, target_path: String) -> Result<()> {
-    Ok(op.copy(&source_path, &target_path).await?)
-}
-
 /// # Safety
 ///
 /// This function should not be called before the Operator is ready.
@@ -463,17 +496,17 @@ fn intern_rename(
     let target_path = jstring_to_string(env, &target_path)?;
 
     executor_or_default(env, executor)?.spawn(async move {
-        let result = do_rename(op, source_path, target_path).await;
-        complete_future(id, result.map(|_| JValueOwned::Void))
+        let result = op
+            .rename(&source_path, &target_path)
+            .await
+            .map(|_| JValueOwned::Void)
+            .map_err(Into::into);
+        complete_future(id, result)
     });
 
     Ok(id)
 }
 
-async fn do_rename(op: &mut Operator, source_path: String, target_path: String) -> Result<()> {
-    Ok(op.rename(&source_path, &target_path).await?)
-}
-
 /// # Safety
 ///
 /// This function should not be called before the Operator is ready.
@@ -503,17 +536,17 @@ fn intern_remove_all(
     let path = jstring_to_string(env, &path)?;
 
     executor_or_default(env, executor)?.spawn(async move {
-        let result = do_remove_all(op, path).await;
-        complete_future(id, result.map(|_| JValueOwned::Void))
+        let result = op
+            .remove_all(&path)
+            .await
+            .map(|_| JValueOwned::Void)
+            .map_err(Into::into);
+        complete_future(id, result)
     });
 
     Ok(id)
 }
 
-async fn do_remove_all(op: &mut Operator, path: String) -> Result<()> {
-    Ok(op.remove_all(&path).await?)
-}
-
 /// # Safety
 ///
 /// This function should not be called before the Operator is ready.
@@ -524,8 +557,9 @@ pub unsafe extern "system" fn Java_org_apache_opendal_AsyncOperator_list(
     op: *mut Operator,
     executor: *const Executor,
     path: JString,
+    options: JObject,
 ) -> jlong {
-    intern_list(&mut env, op, executor, path).unwrap_or_else(|e| {
+    intern_list(&mut env, op, executor, path, options).unwrap_or_else(|e| {
         e.throw(&mut env);
         0
     })
@@ -536,32 +570,38 @@ fn intern_list(
     op: *mut Operator,
     executor: *const Executor,
     path: JString,
+    options: JObject,
 ) -> Result<jlong> {
     let op = unsafe { &mut *op };
     let id = request_id(env)?;
 
     let path = jstring_to_string(env, &path)?;
+    let recursive = read_bool_field(env, &options, "recursive")?;
+
+    let mut list_op = op.list_with(&path);
+    list_op = list_op.recursive(recursive);
 
     executor_or_default(env, executor)?.spawn(async move {
-        let result = do_list(op, path).await;
+        let entries = list_op.await.map_err(Into::into);
+        let result = make_entries(entries);
         complete_future(id, result.map(JValueOwned::Object))
     });
 
     Ok(id)
 }
 
-async fn do_list<'local>(op: &mut Operator, path: String) -> Result<JObject<'local>> {
-    let obs = op.list(&path).await?;
+fn make_entries<'local>(entries: Result<Vec<Entry>>) -> Result<JObject<'local>> {
+    let entries = entries?;
 
     let mut env = unsafe { get_current_env() };
     let jarray = env.new_object_array(
-        obs.len() as jsize,
+        entries.len() as jsize,
         "org/apache/opendal/Entry",
         JObject::null(),
     )?;
 
-    for (idx, entry) in obs.iter().enumerate() {
-        let entry = make_entry(&mut env, entry.to_owned())?;
+    for (idx, entry) in entries.into_iter().enumerate() {
+        let entry = make_entry(&mut env, entry)?;
         env.set_object_array_element(&jarray, idx as jsize, entry)?;
     }
 
