@@ -27,7 +27,7 @@ use jni::sys::jsize;
 use jni::JNIEnv;
 use opendal::BlockingOperator;
 
-use crate::convert::jstring_to_string;
+use crate::convert::{jstring_to_string, read_bool_field, read_map_field, read_string_field};
 use crate::make_entry;
 use crate::make_metadata;
 use crate::Result;
@@ -90,8 +90,9 @@ pub unsafe extern "system" fn Java_org_apache_opendal_Operator_write(
     op: *mut BlockingOperator,
     path: JString,
     content: JByteArray,
+    write_options: JObject,
 ) {
-    intern_write(&mut env, &mut *op, path, content).unwrap_or_else(|e| {
+    intern_write(&mut env, &mut *op, path, content, write_options).unwrap_or_else(|e| {
         e.throw(&mut env);
     })
 }
@@ -101,10 +102,32 @@ fn intern_write(
     op: &mut BlockingOperator,
     path: JString,
     content: JByteArray,
+    options: JObject,
 ) -> Result<()> {
     let path = jstring_to_string(env, &path)?;
     let content = env.convert_byte_array(content)?;
-    Ok(op.write(&path, content).map(|_| ())?)
+
+    let content_type = read_string_field(env, &options, "contentType")?;
+    let content_disposition = read_string_field(env, &options, "contentDisposition")?;
+    let cache_control = read_string_field(env, &options, "cacheControl")?;
+    let user_metadata = read_map_field(env, &options, "userMetadata")?;
+    let append = read_bool_field(env, &options, "append")?;
+
+    let mut write_op = op.write_with(&path, content);
+    if let Some(content_type) = content_type {
+        write_op = write_op.content_type(&content_type);
+    }
+    if let Some(content_disposition) = content_disposition {
+        write_op = write_op.content_disposition(&content_disposition);
+    }
+    if let Some(cache_control) = cache_control {
+        write_op = write_op.cache_control(&cache_control);
+    }
+    if let Some(user_metadata) = user_metadata {
+        write_op = write_op.user_metadata(user_metadata);
+    }
+    write_op = write_op.append(append);
+    Ok(write_op.call().map(|_| ())?)
 }
 
 /// # Safety
@@ -255,25 +278,36 @@ pub unsafe extern "system" fn Java_org_apache_opendal_Operator_list(
     _: JClass,
     op: *mut BlockingOperator,
     path: JString,
+    options: JObject,
 ) -> jobjectArray {
-    intern_list(&mut env, &mut *op, path).unwrap_or_else(|e| {
+    intern_list(&mut env, &mut *op, path, options).unwrap_or_else(|e| {
         e.throw(&mut env);
         JObject::default().into_raw()
     })
 }
 
-fn intern_list(env: &mut JNIEnv, op: &mut BlockingOperator, path: JString) -> Result<jobjectArray> {
+fn intern_list(
+    env: &mut JNIEnv,
+    op: &mut BlockingOperator,
+    path: JString,
+    options: JObject,
+) -> Result<jobjectArray> {
     let path = jstring_to_string(env, &path)?;
-    let obs = op.list(&path)?;
+    let recursive = read_bool_field(env, &options, "recursive")?;
+
+    let mut list_op = op.list_with(&path);
+    list_op = list_op.recursive(recursive);
+
+    let entries = list_op.call()?;
 
     let jarray = env.new_object_array(
-        obs.len() as jsize,
+        entries.len() as jsize,
         "org/apache/opendal/Entry",
         JObject::null(),
     )?;
 
-    for (idx, entry) in obs.iter().enumerate() {
-        let entry = make_entry(env, entry.to_owned())?;
+    for (idx, entry) in entries.into_iter().enumerate() {
+        let entry = make_entry(env, entry)?;
         env.set_object_array_element(&jarray, idx as jsize, entry)?;
     }
 
